use alloc::rc::Rc;

use crate::app_interface::list::ImmutableList;
use crate::app_interface::BattleSideImpl;
use crate::app_interface::{
    BattleParty, BattlePartyImpl, BattleRandom, ChoiceQueue, DynamicLibrary, Pokemon, StringKey,
};

#[cfg_attr(feature = "mock_data", mockall::automock)]
pub trait BattleTrait {
    fn library(&self) -> DynamicLibrary;
    fn parties(&self) -> ImmutableList<BattlePartyImpl>;
    fn sides(&self) -> ImmutableList<BattleSideImpl>;
    fn random(&self) -> BattleRandom;
    fn choice_queue(&self) -> ChoiceQueue;
    fn get_pokemon(&self, side: u8, index: u8) -> Option<Pokemon>;
    fn find_party_for_pokemon(&self, pokemon: &Pokemon) -> Option<BattleParty>;
    fn weather_name(&self) -> Option<StringKey>;
    fn has_weather(&self, name: &str) -> bool;
    fn can_flee(&self) -> bool;
    fn number_of_sides(&self) -> u8;
    fn pokemon_per_side(&self) -> u8;
    fn has_ended(&self) -> bool;
    fn has_ended_conclusively(&self) -> bool;
    fn winning_side(&self) -> u8;
    fn current_turn(&self) -> u32;
}

pub type Battle = Rc<dyn BattleTrait>;
#[cfg(feature = "mock_data")]
pub type MockBattle = MockBattleTrait;

#[cfg(not(feature = "mock_data"))]
mod implementation {
    use super::*;
    use crate::app_interface::PokemonImpl;
    use crate::app_interface::{
        BattleParty, BattlePartyImpl, BattleRandom, BattleRandomImpl, BattleSide, BattleSideImpl,
        ChoiceQueue, ChoiceQueueImpl, Pokemon,
    };
    use crate::handling::cached_value::CachedValue;
    use crate::handling::Cacheable;
    use crate::{
        cached_value, cached_value_getters, wasm_value_getters, wasm_value_getters_extern,
        wasm_value_getters_funcs, DynamicLibrary, ExternRef, ExternalReferenceType, ImmutableList,
        StringKey, VecExternRef,
    };

    struct BattleInner {
        reference: ExternRef<BattleImpl>,
        library: CachedValue<DynamicLibrary>,
        parties: CachedValue<ImmutableList<BattlePartyImpl>>,
        sides: CachedValue<ImmutableList<BattleSideImpl>>,
        random: CachedValue<Rc<BattleRandomImpl>>,
        choice_queue: CachedValue<Rc<ChoiceQueueImpl>>,
    }

    #[derive(Clone)]
    pub struct BattleImpl {
        inner: Rc<BattleInner>,
    }

    #[cfg(not(feature = "mock_data"))]
    impl BattleImpl {
        pub fn new(reference: ExternRef<BattleImpl>) -> Self {
            Self::from_ref(reference, &|reference| Self {
                inner: Rc::new(BattleInner {
                    reference,
                    library: cached_value!({ battle_get_library(reference).get_value().unwrap() }),
                    parties: cached_value!({ battle_get_parties(reference).get_immutable_list() }),
                    sides: cached_value!({ battle_get_sides(reference).get_immutable_list() }),
                    random: cached_value!({
                        Rc::new(battle_get_random(reference).get_value().unwrap())
                    }),
                    choice_queue: cached_value!({
                        Rc::new(battle_get_choice_queue(reference).get_value().unwrap())
                    }),
                }),
            })
        }
    }

    #[cfg(not(feature = "mock_data"))]
    impl BattleTrait for BattleImpl {
        cached_value_getters! {
            fn library(&self) -> DynamicLibrary;
            fn parties(&self) -> ImmutableList<BattlePartyImpl>;
            fn sides(&self) -> ImmutableList<BattleSideImpl>;
            fn random(&self) -> BattleRandom;
            fn choice_queue(&self) -> ChoiceQueue;
        }

        fn get_pokemon(&self, side: u8, index: u8) -> Option<Pokemon> {
            unsafe {
                let v = battle_get_pokemon(self.inner.reference, side, index).get_value();
                if let Some(v) = v {
                    Some(Rc::new(v))
                } else {
                    None
                }
            }
        }

        fn find_party_for_pokemon(&self, pokemon: &Pokemon) -> Option<BattleParty> {
            unsafe {
                let b =
                    battle_find_party_for_pokemon(self.inner.reference, pokemon.reference().into())
                        .get_value();
                if let Some(b) = b {
                    Some(Rc::new(b))
                } else {
                    None
                }
            }
        }

        fn weather_name(&self) -> Option<StringKey> {
            unsafe { battle_get_weather_name(self.inner.reference).get_value() }
        }
        fn has_weather(&self, name: &str) -> bool {
            if let Some(weather) = self.weather_name() {
                if weather.eq(name) {
                    return true;
                }
            }
            false
        }

        wasm_value_getters_funcs! {
            Battle,
            fn can_flee(&self) -> bool;
            fn number_of_sides(&self) -> u8;
            fn pokemon_per_side(&self) -> u8;
            fn has_ended(&self) -> bool;
            fn has_ended_conclusively(&self) -> bool;
            fn winning_side(&self) -> u8;
            fn current_turn(&self) -> u32;
        }
    }

    wasm_value_getters_extern! {
        BattleImpl, Battle,
        pub fn can_flee(&self) -> bool;
        pub fn number_of_sides(&self) -> u8;
        pub fn pokemon_per_side(&self) -> u8;
        pub fn has_ended(&self) -> bool;
        pub fn has_ended_conclusively(&self) -> bool;
        pub fn winning_side(&self) -> u8;
        pub fn current_turn(&self) -> u32;
    }

    crate::handling::cacheable::cacheable!(BattleImpl);

    #[cfg(not(feature = "mock_data"))]
    impl ExternalReferenceType for BattleImpl {
        fn from_extern_value(reference: ExternRef<Self>) -> Self {
            Self::new(reference)
        }
    }

    #[cfg(not(feature = "mock_data"))]
    extern "wasm" {
        fn battle_get_library(r: ExternRef<BattleImpl>) -> ExternRef<DynamicLibrary>;
        fn battle_get_parties(r: ExternRef<BattleImpl>) -> VecExternRef<BattlePartyImpl>;
        fn battle_get_sides(r: ExternRef<BattleImpl>) -> VecExternRef<BattleSideImpl>;
        fn battle_get_random(r: ExternRef<BattleImpl>) -> ExternRef<BattleRandomImpl>;
        fn battle_get_choice_queue(r: ExternRef<BattleImpl>) -> ExternRef<ChoiceQueueImpl>;
        fn battle_get_pokemon(
            r: ExternRef<BattleImpl>,
            side: u8,
            index: u8,
        ) -> ExternRef<PokemonImpl>;
        fn battle_find_party_for_pokemon(
            r: ExternRef<BattleImpl>,
            mon: ExternRef<PokemonImpl>,
        ) -> ExternRef<BattlePartyImpl>;

        fn battle_get_weather_name(r: ExternRef<BattleImpl>) -> ExternRef<StringKey>;
    }
}

#[cfg(not(feature = "mock_data"))]
pub use implementation::*;
